from .canvas import Canvas
from .core import Shape, Bounds, Point, deg
from .base import (
    Rect,
    Square,
    Circle,
    Ellipse,
    Line,
    Arrow,
    Group,
    Path,
    Polyline,
    Text,
)

__version__ = "0.8.1"
